// Copyright Krzysztof Czainski 2011
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

/*
 * @file ptr.hpp
 *
 * @since 14-07-2012
 * @author Krzysztof Czainski
 */

#ifndef PTR_HPP_
#define PTR_HPP_

#include <boost/assert.hpp>
#include <boost/checked_delete.hpp>
#include <boost/move/move.hpp>
#include <boost/swap.hpp>
#include <boost/type_traits/is_convertible.hpp>
#include <boost/utility/enable_if.hpp>

#include <iostream>
#include <typeinfo>

extern int ptr_s;

template < class T >
T* clone_me( T const* p )
{
    if ( p && typeid(T) != typeid(*p) )
        std::cout << "slice! ";
    return p ? new T(*p) : 0;
}

template < class T >
class ptr
{
public:

    explicit ptr( T* p = 0 ) : p_(p) { ++ptr_s; std::cout << "_p "; }

    // copy ctor
    ptr( ptr const& b ) : p_( clone_me(b.get()) ) { ++ptr_s; std::cout << "cp "; }

    // move ctor
    ptr( BOOST_RV_REF(ptr) b ) : p_( b.release() ) { ++ptr_s; std::cout << "mp "; }

private:

#if !defined(BY_VALUE)

    BOOST_COPYABLE_AND_MOVABLE(ptr)

public:

    // generalized copy ctor for pointers to derived
    template < class U >
    ptr( ptr<U> const& b,
            typename boost::enable_if< boost::is_convertible<U*,T*> >::type* = 0 )
    : p_( clone_me(b.get()) ) { ++ptr_s; std::cout << "Cp "; }

    // generalized move ctor for pointers to derived
    template < class U >
    ptr( BOOST_RV_REF(ptr<U>) b,
            typename boost::enable_if< boost::is_convertible<U*,T*> >::type* = 0 )
    : p_( b.release() ) { ++ptr_s; std::cout << "Mp "; }

    ptr& operator=( BOOST_COPY_ASSIGN_REF(ptr) b )
    {
        T* tmp = clone_me(b.get()); // this can throw
        boost::checked_delete(p_);
        p_ = tmp;
        return *this;
    }

    ptr& operator=( BOOST_RV_REF(ptr) b )
    {
        boost::checked_delete(p_);
        p_ = b.release();
        return *this;
    }

    template < class U >
    typename boost::enable_if< boost::is_convertible<U*,T*>, ptr& >::type
    operator=( ptr<U> const& b )
    {
        T* tmp = clone_me(b.get()); // this could throw
        boost::checked_delete(p_);
        p_ = tmp;
        return *this;
    }

    template < class U >
    typename boost::enable_if< boost::is_convertible<U*,T*>, ptr& >::type
    operator=( BOOST_RV_REF(ptr<U>) b )
    {
        boost::checked_delete(p_);
        p_ = b.release();
        return *this;
    }

#else // BY_VALUE

    BOOST_COPYABLE_AND_MOVABLE_ALT(ptr)

public:

    // generalized copy/move constructor implemented by pass-by-value & steal
    template < class U >
    ptr( ptr<U> b,
            typename boost::enable_if< boost::is_convertible<U*,T*> >::type* = 0 )
    : p_( b.release() ) { ++ptr_s; std::cout << "Vp "; }

    // assignment - works for all types convertible to ptr
    ptr& operator=( ptr b )
    {
        swap(*this,b);
        return *this;
    }

#endif // BY_VALUE

    ~ptr() { boost::checked_delete(p_); ++ptr_s; std::cout << "~p "; }

    T* get() const { return p_; }

    T* release() { T* r = p_; p_ = 0; return r; }

    friend void swap( ptr& a, ptr& b )
    {
        boost::swap(a.p_,b.p_);
        std::cout << "sp ";
    }

    // reset, operator* and ->

private:

    T* p_;
};

#endif // PTR_HPP_
